from enum import Enum
from pathlib import Path
from typing import Callable

__all__ = [
    'Reader',
    'Speed',
    'Transcriber',
]


class Speed(Enum):
    SLOW = 0.5
    NORMAL = 1.0


type Reader = Callable[[Speed], None]
type Transcriber = Callable[[Path], str]



from dataclasses import dataclass
from functools import partial
from typing import Callable, cast, Iterable

from rich.text import Text

from reling.config import MAX_SCORE
from reling.db.models import DialogueExam, DialogueExamResult, TextExam, TextExamResult
from reling.helpers.colors import fade
from reling.helpers.diff import highlight_diff
from reling.helpers.output import output, SentenceData
from reling.helpers.scoring import format_average_score
from reling.helpers.wave import play
from reling.tts import TTSVoiceClient
from reling.types import DialogueExchangeData, Input
from reling.utils.console import stream_print_markdown
from reling.utils.iterables import extract_items
from reling.utils.prompts import PROMPT_SEPARATOR, PromptOption
from reling.utils.time import format_time_delta
from reling.utils.transformers import get_numbering_prefix
from .types import ExchangeWithTranslation, ExplanationRequest, SentenceWithTranslation

__all__ = [
    'present_results',
]

NOTHING_TO_IMPROVE = fade('(no changes needed)')


@dataclass
class TitleData:
    text: str
    should_number: bool
    tts: TTSVoiceClient | None


def format_provided_and_suggestion(
        provided: str | None,
        perfect: str | None,
) -> tuple[str | Text | None, str | Text | None]:
    """Compute the representation of the provided answer and suggested improvement to be shown to the user."""
    if provided is not None and perfect is not None:
        if provided == perfect:
            return provided, NOTHING_TO_IMPROVE
        else:
            return highlight_diff(provided, perfect)
    else:
        return provided, perfect


def build_explanation_printer(
        explain: Callable[[ExplanationRequest], Iterable[str]],
        request: ExplanationRequest,
) -> Callable[[], None]:
    """Create a function to print an explanation."""
    return lambda: stream_print_markdown(explain(request), start=PROMPT_SEPARATOR + '\n')


def do_present(
        titles: list[list[TitleData]],
        provided_translations: list[Input | None],
        original_translations: list[str],
        exam: TextExam | DialogueExam,
        target_tts: TTSVoiceClient | None,
        explain: Callable[[ExplanationRequest], Iterable[str]],
) -> None:
    """Present the results of scoring translations."""
    result_indices = [result.index for result in exam.results]
    for title_items, provided_translation, original_translation, result in zip(
            extract_items(titles, result_indices),
            extract_items(provided_translations, result_indices),
            extract_items(original_translations, result_indices),
            cast(list[TextExamResult] | list[DialogueExamResult], exam.results),
    ):
        for title in title_items:
            output(SentenceData.from_tts(
                title.text,
                title.tts,
                print_prefix=get_numbering_prefix(result.index) if title.should_number else '',
            ))
        print(f'Your score: {result.score}/{MAX_SCORE}')
        provided_text = provided_translation.text or None
        perfect_text = (((result.suggested_answer if result.suggested_answer != provided_text else None)
                         or (provided_text if result.score == MAX_SCORE else None)) if provided_text else None)
        provided_print_text, improved_print_text = format_provided_and_suggestion(provided_text, perfect_text)
        output(*[
            SentenceData(
                print_text=provided_print_text,
                print_prefix='Provided: ',
                reader=partial(play, provided_translation.audio) if provided_translation.audio and target_tts else None,
                reader_id='provided',
            ),
            SentenceData.from_tts(
                text=perfect_text,
                client=target_tts,
                print_text=improved_print_text,
                print_prefix='Improved: ',
                reader_id='improved',
            ),
            SentenceData.from_tts(
                text=original_translation,
                client=target_tts,
                print_prefix='Original: ',
                reader_id='original',
            ),
        ], extra_options=[
            PromptOption(
                'explain',
                build_explanation_printer(explain, ExplanationRequest(result.index, source=False)),
                {'source': build_explanation_printer(explain, ExplanationRequest(result.index, source=True))},
            ),
        ])
        print()
    print(f'Average score: {format_average_score(exam)}')
    print(f'Exam duration: {format_time_delta(exam.duration)}')


def present_results(
        items: list[SentenceWithTranslation | ExchangeWithTranslation],
        original_translations: list[str | DialogueExchangeData],
        exam: TextExam | DialogueExam,
        source_tts: TTSVoiceClient | None,
        target_tts: TTSVoiceClient | None,
        target_speaker_tts: TTSVoiceClient | None,
        explain: Callable[[ExplanationRequest], Iterable[str]],
) -> None:
    """Present the results of scoring text or dialogue translations."""
    is_dialogue = isinstance(exam, DialogueExam)
    do_present(
        titles=[([
            TitleData(
                text=original_translation.speaker,
                should_number=False,
                tts=target_speaker_tts,
            ),
        ] if is_dialogue else []) + [
            TitleData(
                text=item.exchange.user if is_dialogue else item.sentence,
                should_number=True,
                tts=source_tts,
            ),
        ] for original_translation, item in zip(
            original_translations,
            items,
        )],
        provided_translations=[item.input for item in items],
        original_translations=([exchange.user for exchange in original_translations]
                               if is_dialogue
                               else original_translations),
        exam=exam,
        target_tts=target_tts,
        explain=explain,
    )



from __future__ import annotations

from lcs2 import lcs_length

SHARED_THRESHOLD = 0.61


class FuzzyWord:
    _normalized: str

    def __init__(self, word: str) -> None:
        self._normalized = word.lower()

    @staticmethod
    def compare(a: FuzzyWord, b: FuzzyWord) -> float:
        """
        Return the similarity score between two FuzzyWord instances.
        Two words are considered more similar if, when lowercased, they share a greater number
        of characters and have a higher ratio of shared characters.
        """
        if not a._normalized and not b._normalized:
            return 1.0
        shared_count = lcs_length(a._normalized, b._normalized)
        shared_ratio = 2 * shared_count / (len(a._normalized) + len(b._normalized))
        return shared_count * shared_ratio if shared_ratio >= SHARED_THRESHOLD else 0.0



from .dialogues import Dialogue, DialogueExam, DialogueExamResult, DialogueExchange, DialogueExchangeTranslation
from .grammar import GrammarCacheSentence, GrammarCacheWord
from .languages import Language
from .misc import IdIndex
from .modifiers import Speaker, Style, Topic
from .texts import Text, TextExam, TextExamResult, TextSentence, TextSentenceTranslation

__all__ = [
    'Dialogue',
    'DialogueExam',
    'DialogueExamResult',
    'DialogueExchange',
    'DialogueExchangeTranslation',
    'GrammarCacheSentence',
    'GrammarCacheWord',
    'IdIndex',
    'Language',
    'Speaker',
    'Style',
    'Text',
    'TextExam',
    'TextExamResult',
    'TextSentence',
    'TextSentenceTranslation',
    'Topic',
]



from dataclasses import dataclass
from pathlib import Path
import sys
from typing import Iterable

from reling.db import init_db as do_init_db, migrate

__all__ = [
    'DatabaseVersion',
    'get_db_path',
    'init_db',
]

DB_PATH: Path | None = None


@dataclass
class DatabaseVersion:
    version: str
    path: Path


def try_migrate(versions: Iterable[DatabaseVersion]) -> DatabaseVersion:
    """
    Attempt to migrate the database to the latest version by iterating through the provided database versions
    to find an existing database file. If found, perform step-by-step migration to the latest version.

    :param versions: An iterable of `DatabaseVersion` objects, ordered from latest to oldest.
    :return: The `DatabaseVersion` object representing the latest version, even if no database file was found.
    """
    iter_versions = iter(versions)
    latest_version = next(iter_versions)
    if not latest_version.path.exists():
        versions_to_upgrade: list[DatabaseVersion] = []
        while True:
            if not (version := next(iter_versions, None)):
                break
            versions_to_upgrade.append(version)
            if version.path.exists():
                for from_version, to_version in zip(
                        reversed(versions_to_upgrade),
                        reversed([latest_version] + versions_to_upgrade[:-1]),
                ):
                    migrate(from_version.path, from_version.version)
                    from_version.path.rename(to_version.path)
                    print(f'Migrated database from version {from_version.version} to {to_version.version}',
                          file=sys.stderr)
    return latest_version


def init_db(versions: Iterable[DatabaseVersion]) -> None:
    """
    Initialize the database by finding or creating the latest version.

    This function first checks whether the database has already been initialized. If not, it attempts to
    migrate an existing database file to the latest version by calling `try_migrate`. If no existing
    database file is found, it initializes a new database at the latest version.

    :param versions: An iterable of `DatabaseVersion` objects, ordered from latest to oldest.
    :raises RuntimeError: If the database has already been initialized.
    """
    global DB_PATH
    if DB_PATH is not None:
        raise RuntimeError('Database is already initialized.')
    DB_PATH = try_migrate(versions).path
    do_init_db(f'sqlite:///{DB_PATH}')


def get_db_path() -> Path:
    if DB_PATH is None:
        raise RuntimeError('Database is not initialized.')
    return DB_PATH



from __future__ import annotations
from abc import ABC, abstractmethod

from reling.types import Reader, Speed
from .voices import Voice

__all__ = [
    'TTSClient',
    'TTSVoiceClient',
]


class TTSClient(ABC):
    @abstractmethod
    def read(self, text: str, voice: Voice, speed: Speed) -> None:
        """Read the text in real time using the specified voice."""
        pass

    def with_voice(self, voice: Voice) -> TTSVoiceClient:
        return TTSVoiceClient(self, voice)


class TTSVoiceClient:
    """A wrapper around TTSClient with a specific voice."""
    _tts: TTSClient
    _voice: Voice

    def __init__(self, tts: TTSClient, voice: Voice) -> None:
        self._tts = tts
        self._voice = voice

    def get_reader(self, text: str) -> Reader:
        def read(speed: Speed) -> None:
            self._tts.read(text, self._voice, speed)
        return read



from sqlalchemy.exc import IntegrityError

from reling.db import single_session
from reling.db.models import Dialogue, DialogueExchangeTranslation, Language, Text, TextSentenceTranslation
from reling.types import DialogueExchangeData
from .exceptions import TranslationExistsException


__all__ = [
    'save_dialogue_translation',
    'save_text_translation',
]


def save_text_translation(
        source_text: Text,
        target_language: Language,
        translated_sentences: list[str],
) -> None:
    """
    Save the translation of a text.
    :raises TranslationExistsError: If the text has already been translated into the target language.
    """
    try:
        with single_session() as session:
            session.add_all([
                TextSentenceTranslation(
                    text_id=source_text.id,
                    language_id=target_language.id,
                    text_sentence_index=index,
                    sentence=sentence,
                )
                for index, sentence in enumerate(translated_sentences)
            ])
            session.commit()
    except IntegrityError:
        raise TranslationExistsException


def save_dialogue_translation(
        source_dialogue: Dialogue,
        target_language: Language,
        translated_exchanges: list[DialogueExchangeData],
) -> None:
    """
    Save the translation of a dialogue.
    :raises TranslationExistsError: If the dialogue has already been translated into the target language.
    """
    try:
        with single_session() as session:
            session.add_all([
                DialogueExchangeTranslation(
                    dialogue_id=source_dialogue.id,
                    language_id=target_language.id,
                    dialogue_exchange_index=index,
                    speaker=exchange.speaker,
                    user=exchange.user,
                )
                for index, exchange in enumerate(translated_exchanges)
            ])
            session.commit()
    except IntegrityError:
        raise TranslationExistsException



from __future__ import annotations
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from math import floor, log2

from reling.db import single_session
from reling.db.models import Dialogue, Language, Text
from reling.utils.tables import build_table, print_table
from reling.utils.time import now
from .streaks import compute_streaks, LanguagePair, Streak

__all__ = [
    'compute_repetition_data',
    'print_repetition_statistics',
    'RepetitionData',
    'RepetitionCategoryStatistics',
    'RepetitionStatistics',
]

MIN_WAIT_PERIOD = timedelta(hours=1)

SESSION_SUMMARY_TITLE = 'Session Summary'
CUMULATIVE_REVIEWS_TITLE = 'Estimated Cumulative Sentence Reviews'

METRIC = 'Metric'
TEXTS = 'Texts'
DIALOGUES = 'Dialogues'
TOTAL = 'Total'

EXAMS = 'Exams to take'
SENTENCES = 'Sentences to review'

PERIOD = 'Period'

PERIODS: dict[int, str] = {
    1: 'day',
    7: 'week',
    30: 'month',
    365: 'year',
}

COLUMN_WIDTH = max(len(METRIC), len(PERIOD), len(TEXTS), len(DIALOGUES), len(TOTAL))


@dataclass
class RepetitionCategoryStatistics:
    exams: int = 0
    sentences: int = 0
    cumulative_reviews: dict[int, int] = field(default_factory=lambda: {period: 0 for period in PERIODS})


@dataclass
class RepetitionStatistics:
    texts: RepetitionCategoryStatistics = field(default_factory=RepetitionCategoryStatistics)
    dialogues: RepetitionCategoryStatistics = field(default_factory=RepetitionCategoryStatistics)


@dataclass
class ExamInfo:
    content: Text | Dialogue
    source_language: Language
    target_language: Language
    skipped_indices: set[int]

    @property
    def sentences_to_review(self) -> int:
        return self.content.size - len(self.skipped_indices)

    def __gt__(self, other: ExamInfo) -> bool:
        return self.sentences_to_review > other.sentences_to_review or (
            self.sentences_to_review == other.sentences_to_review and
            self.content.created_at < other.content.created_at
        )


@dataclass
class RepetitionData:
    statistics: RepetitionStatistics = field(default_factory=RepetitionStatistics)
    next_exam: ExamInfo | None = None


def get_next_review_times(streak: Streak, reference_time: datetime) -> tuple[datetime, timedelta, timedelta]:
    """
    Compute the next scheduled review time and wait periods for subsequent reviews,
    assuming all reviews are successful and occur on time.
    """
    if streak.timeline:
        wait_period = max(streak.timeline.last - streak.timeline.first, MIN_WAIT_PERIOD)
        next_review_time = max(streak.timeline.last + wait_period, reference_time)
        next_review_wait = next_review_time - streak.timeline.first
        return next_review_time, next_review_wait, 2 * next_review_wait
    else:
        return reference_time, MIN_WAIT_PERIOD, MIN_WAIT_PERIOD


def get_num_reviews_before(
        first_review: datetime,
        first_wait: timedelta,
        second_wait: timedelta,
        before: datetime,
) -> int:
    """
    Return the total number of on-time, successful reviews that occur strictly before a given time;
    starting from the third review, the wait time doubles after each successful review.
    """
    if first_review >= before:
        return 0
    second_review = first_review + first_wait
    if second_review >= before:
        return 1
    return floor(log2((before - second_review) // second_wait + 1)) + 2


def update(
        data: RepetitionData,
        content: Text | Dialogue,
        streaks: list[Streak],
        languages: LanguagePair,
        reference_time: datetime,
) -> None:
    """Update the repetition data given the streaks for a content."""
    stats = data.statistics.texts if isinstance(content, Text) else data.statistics.dialogues
    skipped_indices: set[int] = set()
    for index, streak in enumerate(streaks):
        first_review, first_wait, second_wait = get_next_review_times(streak, reference_time)
        if first_review == reference_time:
            stats.sentences += 1
        else:
            skipped_indices.add(index)
        for period in PERIODS:
            stats.cumulative_reviews[period] += get_num_reviews_before(
                first_review,
                first_wait,
                second_wait,
                reference_time + timedelta(days=period),
            )
    candidate = ExamInfo(content, languages.source_language, languages.target_language, skipped_indices)
    if candidate.sentences_to_review > 0:
        stats.exams += 1
        if data.next_exam is None or candidate > data.next_exam:
            data.next_exam = candidate


def compute_repetition_data(source_language: Language | None, target_language: Language | None) -> RepetitionData:
    """Compute spaced repetition mode data."""
    reference_time = now()
    data = RepetitionData()
    with single_session() as session:
        for model in [Text, Dialogue]:
            for content in session.query(model).filter(model.archived_at.is_not(None)):
                for languages, streaks in compute_streaks(content, source_language, target_language).items():
                    update(data, content, streaks, languages, reference_time)
    return data


def print_session_summary_statistics(statistics: RepetitionStatistics) -> None:
    """Display current session statistics."""
    table = build_table(
        title=SESSION_SUMMARY_TITLE,
        headers=[
            METRIC,
            TEXTS,
            DIALOGUES,
            TOTAL,
        ],
        justify={
            METRIC: 'left',
            TEXTS: 'right',
            DIALOGUES: 'right',
            TOTAL: 'right',
        },
        widths={
            METRIC: COLUMN_WIDTH,
            TEXTS: COLUMN_WIDTH,
            DIALOGUES: COLUMN_WIDTH,
            TOTAL: COLUMN_WIDTH,
        },
        data=[
            {
                METRIC: EXAMS,
                TEXTS: str(statistics.texts.exams),
                DIALOGUES: str(statistics.dialogues.exams),
                TOTAL: str(statistics.texts.exams + statistics.dialogues.exams),
            },
            {
                METRIC: SENTENCES,
                TEXTS: str(statistics.texts.sentences),
                DIALOGUES: str(statistics.dialogues.sentences),
                TOTAL: str(statistics.texts.sentences + statistics.dialogues.sentences),
            },
        ],
        group_by=[
            METRIC,
        ],
    )
    print_table(table)


def print_cumulative_reviews_statistics(statistics: RepetitionStatistics) -> None:
    """Display total reviews statistics."""
    metric_column_width = max(len(TEXTS), len(DIALOGUES), len(TOTAL))
    table = build_table(
        title=CUMULATIVE_REVIEWS_TITLE,
        headers=[
            PERIOD,
            TEXTS,
            DIALOGUES,
            TOTAL,
        ],
        justify={
            PERIOD: 'left',
            TEXTS: 'right',
            DIALOGUES: 'right',
            TOTAL: 'right',
        },
        widths={
            PERIOD: COLUMN_WIDTH,
            TEXTS: metric_column_width,
            DIALOGUES: metric_column_width,
            TOTAL: metric_column_width,
        },
        data=[{
            PERIOD: period_title.capitalize(),
            TEXTS: str(statistics.texts.cumulative_reviews[period]),
            DIALOGUES: str(statistics.dialogues.cumulative_reviews[period]),
            TOTAL: str(statistics.texts.cumulative_reviews[period] + statistics.dialogues.cumulative_reviews[period]),
        } for period, period_title in PERIODS.items()],
        group_by=[
            PERIOD,
        ],
    )
    print_table(table)


def print_repetition_statistics(statistics: RepetitionStatistics) -> None:
    """Display spaced repetition statistics."""
    print()
    print_session_summary_statistics(statistics)
    print()
    print_cumulative_reviews_statistics(statistics)



from contextlib import contextmanager
from itertools import chain
from shutil import get_terminal_size
from typing import Callable, Generator, Iterable

from prompt_toolkit import PromptSession
from prompt_toolkit.history import InMemoryHistory
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from wcwidth import wcwidth

from .strings import universal_normalize

__all__ = [
    'clear_current_line',
    'input_and_erase',
    'interruptible_input',
    'print_and_erase',
    'print_and_maybe_erase',
    'stream_print_markdown',
]


def clear_current_line() -> None:
    print('\033[2K', end='\r')


def clear_previous(lines: int = 1) -> None:
    print('\033[F\033[K' * lines, end='\r')


def count_lines(text: str, num_columns: int) -> int:
    """Count the number of lines `text` takes up when printed with a maximum width of `num_columns`."""
    lines = 1
    taken = 0
    for char in text:
        if char == '\n':
            lines += 1
            taken = 0
        elif (char_width := wcwidth(char)) > 0:
            if taken + char_width <= num_columns:
                taken += char_width
            else:
                lines += 1
                taken = char_width
    return lines


def erase_previous(text: str) -> None:
    clear_current_line()
    clear_previous(count_lines(text, get_terminal_size().columns))


def interruptible_input(prompt: str, suggestions: list[str] | None = None, mouse_support: bool = False) -> str:
    history = InMemoryHistory()
    for suggestion in suggestions or []:
        if suggestion:
            history.append_string(suggestion)
    session = PromptSession(history=history)
    try:
        user_input = session.prompt(prompt, mouse_support=mouse_support)
    except KeyboardInterrupt:
        erase_previous(prompt)
        raise
    return universal_normalize(user_input)


def input_and_erase(prompt: str) -> str:
    data = interruptible_input(prompt)
    erase_previous(prompt + data + ' ')  # The space represents the rightmost position of the cursor
    return data


@contextmanager
def print_and_erase(text: str) -> Generator[None, None, None]:
    print(text)
    yield
    erase_previous(text)


@contextmanager
def print_and_maybe_erase(text: str) -> Generator[Callable[[], None], None, None]:
    print(text)
    yield lambda: erase_previous(text)


def stream_print_markdown(stream: Iterable[str], start: str = '', end: str = '\n') -> None:
    """Print a stream of Markdown-formatted text."""
    console = Console()
    buffer: list[str] = []
    with Live(console=console) as live:
        for chunk in chain([start], stream, [end]):
            buffer.append(chunk)
            live.update(Markdown(''.join(buffer)))



from random import choice
from tqdm import tqdm
import typer

from reling.app.app import app
from reling.app.default_content import set_default_content
from reling.app.types import (
    API_KEY,
    INCLUDE_OPT,
    LANGUAGE_ARG,
    LEVEL_OPT,
    MODEL,
    SIZE_DIALOGUE_OPT,
    SIZE_TEXT_OPT,
    SPEAKER_GENDER_OPT,
    SPEAKER_OPT,
    STYLE_OPT,
    TOPIC_OPT,
    USER_GENDER,
)
from reling.db.enums import Gender, Level
from reling.db.helpers.modifiers import get_random_modifier
from reling.db.models import Speaker, Style, Topic
from reling.gpt import GPTClient
from reling.helpers.typer import typer_raise
from reling.types import WordWithSense
from .generation import generate_dialogue_exchanges, generate_id, generate_text_sentences
from .storage import save_dialogue, save_text

__all__ = [
    'create',
]

DEFAULT_SIZE_TEXT = 10
DEFAULT_SIZE_DIALOGUE = 10

MIN_SIZE_THRESHOLD = 0.9

create = typer.Typer()
app.add_typer(
    create,
    name='create',
    help='Create a new text or dialogue.',
)


@create.command()
def text(
        api_key: API_KEY,
        model: MODEL,
        language: LANGUAGE_ARG,
        level: LEVEL_OPT = Level.INTERMEDIATE,
        topic: TOPIC_OPT = None,
        style: STYLE_OPT = None,
        size: SIZE_TEXT_OPT = DEFAULT_SIZE_TEXT,
        include: INCLUDE_OPT = None,
) -> None:
    """Create a text and save it to the database."""
    gpt = GPTClient(api_key=api_key.get(), model=model.get())
    topic = topic or get_random_modifier(Topic, level).name
    style = style or get_random_modifier(Style, level).name

    sentences = list(tqdm(
        generate_text_sentences(
            gpt=gpt,
            num_sentences=size,
            language=language,
            level=level,
            topic=topic,
            style=style,
            include=list(map(WordWithSense.parse, include or [])),
        ),
        desc=f'Generating text in {language.name}',
        total=size,
        leave=False,
    ))
    if len(sentences) < round(size * MIN_SIZE_THRESHOLD):
        typer_raise('Failed to generate the text.')

    content = save_text(
        suggested_id=generate_id(gpt, sentences),
        sentences=sentences,
        language=language,
        level=level,
        topic=topic,
        style=style,
    )
    set_default_content(content)
    print(f'Generated text with the following ID:\n{content.id}')


@create.command()
def dialogue(
        api_key: API_KEY,
        model: MODEL,
        user_gender: USER_GENDER,
        language: LANGUAGE_ARG,
        level: LEVEL_OPT = Level.INTERMEDIATE,
        speaker: SPEAKER_OPT = None,
        speaker_gender: SPEAKER_GENDER_OPT = None,
        topic: TOPIC_OPT = None,
        size: SIZE_DIALOGUE_OPT = DEFAULT_SIZE_DIALOGUE,
        include: INCLUDE_OPT = None,
) -> None:
    """Create a dialogue and save it to the database."""
    gpt = GPTClient(api_key=api_key.get(), model=model.get())
    speaker = speaker or get_random_modifier(Speaker, level).name
    speaker_gender = speaker_gender or choice([Gender.MALE, Gender.FEMALE])

    exchanges = list(tqdm(
        generate_dialogue_exchanges(
            gpt=gpt,
            num_exchanges=size,
            language=language,
            level=level,
            user_gender=user_gender,
            speaker=speaker,
            speaker_gender=speaker_gender,
            topic=topic,
            include=list(map(WordWithSense.parse, include or [])),
        ),
        desc=f'Generating dialogue in {language.name}',
        total=size,
        leave=False,
    ))
    if len(exchanges) < round(size * MIN_SIZE_THRESHOLD):
        typer_raise('Failed to generate the dialogue.')

    content = save_dialogue(
        suggested_id=generate_id(gpt, [turn for exchange in exchanges for turn in exchange.all()]),
        exchanges=exchanges,
        language=language,
        level=level,
        speaker=speaker,
        topic=topic,
        speaker_gender=speaker_gender,
        user_gender=user_gender,
    )
    set_default_content(content)
    print(f'Generated dialogue with the following ID:\n{content.id}')



from dataclasses import dataclass

__all__ = [
    'DialogueExchangeData',
]


@dataclass
class DialogueExchangeData:
    speaker: str
    user: str

    @staticmethod
    def assert_speaker_comes_first() -> None:
        pass

    def all(self) -> tuple[str, str]:
        return self.speaker, self.user



from random import shuffle
from typing import cast

from reling.db.enums import Gender
from reling.tts import Voice

__all__ = [
    'pick_voice',
    'pick_voices',
]


def pick_voices(*positions: Gender | None) -> tuple[Voice, ...]:
    """
    Pick random non-repeating voices of the specified genders
    (`None` denoting no gender preference for the given position).

    :raises ValueError: If there are not enough voices to satisfy the requirements.
    """
    pools: dict[Gender | None, list[Voice]] = {None: []}

    for gender in Gender:
        gender_voices = [cast(Voice, voice) for voice in Voice if cast(Voice, voice).gender == gender]

        required_count = positions.count(gender)
        if len(gender_voices) < required_count:
            raise ValueError(f'Not enough voices of {gender}.')

        shuffle(gender_voices)
        pools[cast(Gender, gender)] = gender_voices[:required_count]
        pools[None].extend(gender_voices[required_count:])

    if len(pools[None]) < positions.count(None):
        raise ValueError('Not enough voices.')

    shuffle(pools[None])
    return tuple(pools[position].pop() for position in positions)


def pick_voice(gender: Gender | None = None) -> Voice:
    """Pick a random voice of the specified gender (`None` denoting no gender preference)."""
    return pick_voices(gender)[0]



from typing import Iterable

from tqdm import tqdm

from reling.db.models import DialogueExam, TextExam
from .modalities import Modality

__all__ = [
    'progress',
]


def progress[T](
        iterable: Iterable[T],
        *,
        total: int,
        modality: Modality,
        model: type[TextExam | DialogueExam] | None = None,
) -> Iterable[T]:
    """Display the progress of the given iterable."""
    return tqdm(
        iterable,
        desc=f'Computing {modality.value} stats'
             + (f' for {'texts' if model == TextExam else 'dialogues'}' if model else ''),
        total=total,
        leave=False,
    )



from __future__ import annotations
from contextlib import contextmanager
from dataclasses import dataclass
from time import sleep
from typing import Callable, Generator, Mapping, TYPE_CHECKING

if TYPE_CHECKING:
    from pyaudio import PyAudio, Stream

from reling.utils.values import coalesce

__all__ = [
    'get_audio',
    'get_stream',
    'PyAudioData'
]

OUTPUT_STOP_DELAY_SEC = 0.5  # This helps avoid audio glitches at the end of playback


@dataclass
class PyAudioData:
    audio: PyAudio
    paInt16: int
    paContinue: int
    paFramesPerBufferUnspecified: int


@contextmanager
def get_audio() -> Generator[PyAudioData, None, None]:
    """Dynamically import PyAudio and yield it, closing it afterward."""
    import pyaudio  # Only import this module if audio is used
    audio = pyaudio.PyAudio()
    try:
        yield PyAudioData(
            audio,
            paInt16=pyaudio.paInt16,
            paContinue=pyaudio.paContinue,
            paFramesPerBufferUnspecified=pyaudio.paFramesPerBufferUnspecified,
        )
    finally:
        audio.terminate()


@contextmanager
def get_stream(
        pyaudio: PyAudioData,
        format: int,  # noqa
        channels: int,
        rate: int,
        input: bool = False,  # noqa
        output: bool = False,
        frames_per_buffer: int | None = None,
        stream_callback: Callable[[bytes, int, Mapping[str, float], int], tuple[bytes | None, int]] | None = None,
) -> Generator[Stream, None, None]:
    """Create a PyAudio stream and yield it, closing it afterward."""
    stream = pyaudio.audio.open(
        format=format,
        channels=channels,
        rate=rate,
        input=input,
        output=output,
        frames_per_buffer=coalesce(frames_per_buffer, pyaudio.paFramesPerBufferUnspecified),
        stream_callback=stream_callback,
    )
    try:
        yield stream
    finally:
        if output:
            sleep(OUTPUT_STOP_DELAY_SEC)
        stream.stop_stream()
        stream.close()



from __future__ import annotations
from dataclasses import dataclass
from enum import StrEnum

from .console import input_and_erase

__all__ = [
    'ENTER',
    'enter_to_continue',
    'format_shortcut',
    'Prompt',
    'PROMPT_SEPARATOR',
    'PromptOption',
]

PROMPT_SEPARATOR = '-' * 3

ENTER = 'Enter'
ENTER_TO_CONTINUE = f'{ENTER} to continue'

TITLE_DELIMITER = ': '
MODIFIER_DELIMITER = ' + '
OPTION_DELIMITER = ' | '
PROMPT_SUFFIX = ': '

SHORTCUT_FORMAT = '[{shortcut}]'


def format_shortcut(shortcut: str) -> str:
    return SHORTCUT_FORMAT.format(shortcut=shortcut)


def highlight_initial(text: str) -> str:
    return format_shortcut(text[0]) + text[1:]


def normalize(text: str) -> str:
    return text.strip().lower()


@dataclass
class PromptOption[T]:
    """
    A prompt option with an action and optional modifiers.
    The first letters of the description and modifiers will be used as "shortcuts" for the action.
    """
    description: str
    action: T
    modifiers: dict[str, T] | None = None

    def format_description(self) -> str:
        """Format the description line for display (e.g., '[o]riginal + [s]lowly')."""
        return MODIFIER_DELIMITER.join(
            highlight_initial(option)
            for option in [self.description] + list((self.modifiers or {}).keys())
        )

    def get_shortcut_mapping(self) -> dict[str, T]:
        """Get a mapping of shortcuts to actions."""
        return {
            self.description[0]: self.action,
            **{
                self.description[0] + modifier[0]: modifier_action
                for modifier, modifier_action in (self.modifiers or {}).items()
            },
        }

    def format_shortcuts(self) -> str:
        """Format the shortcuts line for display (e.g., '[o] | [os]')."""
        return OPTION_DELIMITER.join(format_shortcut(shortcut) for shortcut in self.get_shortcut_mapping().keys())


class Prompt[T]:
    """A prompt that allows the user to choose from multiple options."""
    _title: str | None
    _options: list[PromptOption[T]]

    def __init__(self, title: str | None = None) -> None:
        self._title = title
        self._options = []

    @staticmethod
    def from_enum[T: StrEnum](enum: T, title: str | None = None) -> Prompt[T]:
        prompt = Prompt(title)
        for option in enum:
            prompt.add_option(PromptOption(
                description=option.value,
                action=option,
            ))
        return prompt

    def add_option(self, option: PromptOption) -> Prompt[T]:
        self._options.append(option)
        return self

    def _match(self, response: str) -> T | None:
        """
        Match the response to the corresponding action (None is returned if the response is empty).
        :raises ValueError: If the response is not one of the shortcuts or empty.
        """
        normalized_response = normalize(response)
        if normalized_response:
            for option in self._options:
                mapping = option.get_shortcut_mapping()
                if normalized_response in mapping:
                    return mapping[normalized_response]
            raise ValueError(f'Invalid response: {response}')
        else:
            return None

    def prompt(self) -> T | None:
        """
        Prompt the user for a choice and return the corresponding action (or None if the response is empty).
        If the response is invalid, the prompt is repeated until a valid response is given.
        """
        if not self._options:
            enter_to_continue()
            return None
        while True:
            response = input_and_erase(
                '\n'.join([
                    PROMPT_SEPARATOR,
                    (self._title + TITLE_DELIMITER if self._title else '') + OPTION_DELIMITER.join(
                        [option.format_description() for option in self._options] + [ENTER_TO_CONTINUE],
                    ),
                    OPTION_DELIMITER.join(
                        [option.format_shortcuts() for option in self._options] + [ENTER],
                    ) + PROMPT_SUFFIX,
                ]),
            )
            try:
                return self._match(response)
            except ValueError:
                pass


def enter_to_continue() -> None:
    """Prompt the user to press Enter to continue."""
    input_and_erase('\n'.join([
        PROMPT_SEPARATOR,
        ENTER_TO_CONTINUE + PROMPT_SUFFIX,
    ]))



from typing import cast

from reling.app.exceptions import AlgorithmException
from reling.db import single_session
from reling.db.models import Dialogue, DialogueExchangeTranslation, Language, Text, TextSentenceTranslation
from reling.gpt import GPTClient
from reling.helpers.typer import typer_raise
from reling.types import DialogueExchangeData, Promise
from .exceptions import TranslationExistsException
from .operation import translate_dialogue, translate_text

__all__ = [
    'get_dialogue_exchanges',
    'get_text_sentences',
]

NO_TRANSLATIONS = 'No translations found.'


def get_text_sentences(text: Text, language: Language, gpt: Promise[GPTClient] | None = None) -> list[str]:
    """Get the sentences of a text in a specified language."""
    if language.id == text.language_id:
        return [cast(str, sentence.sentence) for sentence in text.sentences]
    if gpt is not None:
        try:
            translate_text(gpt, text, language)
        except TranslationExistsException:
            pass
        except AlgorithmException as e:
            typer_raise(e.msg)
    with single_session() as session:
        return [
            translation.sentence
            for translation in session.query(TextSentenceTranslation)
            .where(TextSentenceTranslation.text_id == text.id)
            .where(TextSentenceTranslation.language_id == language.id)
            .order_by(TextSentenceTranslation.text_sentence_index)
        ] or typer_raise(NO_TRANSLATIONS)


def get_dialogue_exchanges(
        dialogue: Dialogue,
        language: Language,
        gpt: Promise[GPTClient] | None = None,
) -> list[DialogueExchangeData]:
    """Get the exchanges of a dialogue in a specified language."""
    if language.id == dialogue.language_id:
        return [
            DialogueExchangeData(
                speaker=exchange.speaker,
                user=exchange.user,
            )
            for exchange in dialogue.exchanges
        ]
    if gpt is not None:
        try:
            translate_dialogue(gpt, dialogue, language)
        except TranslationExistsException:
            pass
        except AlgorithmException as e:
            typer_raise(e.msg)
    with single_session() as session:
        return [
            DialogueExchangeData(
                speaker=translation.speaker,
                user=translation.user,
            )
            for translation in session.query(DialogueExchangeTranslation)
            .where(DialogueExchangeTranslation.dialogue_id == dialogue.id)
            .where(DialogueExchangeTranslation.language_id == language.id)
            .order_by(DialogueExchangeTranslation.dialogue_exchange_index)
        ] or typer_raise(NO_TRANSLATIONS)



from math import ceil, floor
from statistics import mean

from lcs2 import diff, lcs_indices, lcs_length

from reling.config import MAX_SCORE
from reling.db.models import DialogueExam, TextExam
from reling.utils.strings import tokenize
from .fuzzy_word import FuzzyWord

__all__ = [
    'calculate_diff_score',
    'format_average_score',
    'get_average_score',
]


def calculate_mistake_score(mistakes: int) -> int:
    """Return the score based on the number of mistakes."""
    return floor(MAX_SCORE * ((1 - 1 / MAX_SCORE) ** mistakes))


def calculate_lcs_score(lcs_len: int, a_len: int, b_len: int) -> int:
    """Return the score based on the length of the longest common subsequence of two sequences."""
    if lcs_len == a_len == b_len:
        return MAX_SCORE
    else:
        return ceil(lcs_len / max(a_len, b_len) * (MAX_SCORE - 1))


def calculate_char_diff_score(a: str, b: str) -> int:
    """Return the score based on the longest common subsequence of two strings."""
    return calculate_lcs_score(lcs_length(a, b), len(a), len(b))


def calculate_fuzzy_word_diff_score(a: str, b: str, cj: bool) -> int:
    """
    Calculate the score based on the total length of the longest common subsequences of characters
    within fuzzily aligned words from two strings.
    """
    a_words, b_words = (tokenize(sentence, punctuation=False, whitespace=False, cj=cj) for sentence in (a, b))
    a_len, b_len = (sum(map(len, words)) for words in (a_words, b_words))
    lcs_len = 0
    for a_index, b_index in lcs_indices(map(FuzzyWord, a_words), map(FuzzyWord, b_words), FuzzyWord.compare):
        lcs_len += lcs_length(a_words[a_index], b_words[b_index])
    return calculate_lcs_score(lcs_len, a_len, b_len)


def calculate_word_mistake_diff_score(a: str, b: str, cj: bool) -> int:
    """Return the score based on the number of mistakes computed from a word-level diff between two strings."""
    a_words, b_words = ([word.lower() for word in tokenize(sentence, punctuation=False, whitespace=False, cj=cj)]
                        for sentence in (a, b))
    a_diff: list[str] = []
    b_diff: list[str] = []
    mistakes = 0
    for a_tokens, b_tokens in diff(a_words, b_words):
        mistakes += max(len(a_tokens), len(b_tokens))
        a_diff.extend(a_tokens)
        b_diff.extend(b_tokens)
    mistakes -= lcs_length(a_diff, b_diff)
    return calculate_mistake_score(mistakes)


def calculate_diff_score(a: str, b: str) -> int:
    """Return the score based on the diff between two strings."""
    return min(
        calculate_char_diff_score(a, b),
        calculate_fuzzy_word_diff_score(a, b, cj=True),
        calculate_word_mistake_diff_score(a, b, cj=False),
    )


def get_average_score(exam: TextExam | DialogueExam) -> float:
    """Return the average score of the exam."""
    return mean(result.score for result in exam.results)


def format_average_score(exam: TextExam | DialogueExam) -> str:
    return f'{get_average_score(exam):.1f}'



from sqlalchemy.orm import Mapped, mapped_column

from reling.db.base import Base
from reling.db.enums import Level

__all__ = [
    'Speaker',
    'Style',
    'Topic',
]


class Topic(Base):
    __tablename__ = 'topics'

    name: Mapped[str] = mapped_column(primary_key=True)
    level: Mapped[Level]


class Style(Base):
    __tablename__ = 'styles'

    name: Mapped[str] = mapped_column(primary_key=True)
    level: Mapped[Level]


class Speaker(Base):
    __tablename__ = 'speakers'

    name: Mapped[str] = mapped_column(primary_key=True)
    level: Mapped[Level]



from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from heapq import merge
from typing import cast

from rich.text import Text

from reling.app.translation import get_dialogue_exchanges, get_text_sentences
from reling.config import MAX_SCORE
from reling.db import single_session
from reling.db.models import DialogueExam, Language, TextExam
from reling.helpers.grammar import Analyzer, WordInfo
from reling.utils.console import print_and_erase
from reling.utils.iterables import extract_items
from reling.utils.tables import build_table, print_table
from .checkpoints import colorize, since
from .filter import get_filter
from .modalities import Modality
from .pos import Pos
from .progress import progress
from .table_str_enum import TableStrEnum

__all__ = [
    'display_stats',
]

PREPARING = 'Preparing...'

TITLE = '{modality}\n(only perfect answers are included; entry = unique text or dialogue)'

POS = 'Part of\nspeech'
TYPE = 'Word\ntype'
ONCE = '≥ 1\nentry'
OCCURRENCES = '≥ {occurrences}\nentries'

TOTAL = 'Total'

STATS_TYPE_DIVIDER = '---'

type Threshold = int

REPETITION_THRESHOLDS: list[Threshold] = [1, 3, 10, 25, 100]


class StatsType(TableStrEnum):
    FORMS = 'forms'
    LEMMAS = 'lemmas'


type Stats = defaultdict[tuple[Pos, StatsType, Threshold], int]

type NormalizedForm = tuple[str, str]


@dataclass
class Checkpoint:
    start_time: datetime
    stats: Stats


@dataclass
class PeriodStats:
    all_time: Stats
    checkpoints: list[Checkpoint]


def get_pos_stats(stats: Stats, pos: Pos | None, stats_type: StatsType, threshold: Threshold) -> int:
    """Get the statistics for the given pos, type, and threshold; or, if pos is None, sum over all parts of speech."""
    return (stats[pos, stats_type, threshold]
            if pos else
            sum(stats[cast(Pos, pos_), stats_type, threshold] for pos_ in Pos))


def get_normalized_form(word: WordInfo) -> NormalizedForm:
    return word.lemma, word.text.lower().replace('’', "'")


def get_relevant_periods(stats: PeriodStats, exam: TextExam | DialogueExam) -> list[Stats]:
    """Return the relevant statistics periods for the given exam."""
    return [
        stats.all_time,
        *(checkpoint.stats for checkpoint in stats.checkpoints if checkpoint.start_time <= exam.started_at),
    ]


def update_stats(stats: Stats, pos: Pos, stats_type: StatsType, occurrences: int) -> None:
    """Update the statistics with the given word."""
    for threshold in REPETITION_THRESHOLDS:
        if occurrences == threshold:
            stats[pos, stats_type, threshold] += 1


def get_relevant_sentences(exam: TextExam | DialogueExam, language: Language, modality: Modality) -> list[str]:
    """Return the relevant sentences to analyze."""
    if modality == Modality.PRODUCTION:
        return [result.answer for result in exam.results]
    else:
        return list(extract_items(
            get_text_sentences(exam.text, language)
            if isinstance(exam, TextExam)
            else (exchange.user for exchange in get_dialogue_exchanges(cast(DialogueExam, exam).dialogue, language)),
            (result.index for result in exam.results),
        ))


class StatsHandler:
    stats: PeriodStats
    analyzer: Analyzer
    lemma_content_ids: defaultdict[str, set[str]]
    form_content_ids: defaultdict[NormalizedForm, set[str]]

    def __init__(self, language: Language, checkpoints: list[datetime]) -> None:
        self.stats = PeriodStats(
            all_time=defaultdict(int),
            checkpoints=[Checkpoint(start_time=checkpoint, stats=defaultdict(int)) for checkpoint in checkpoints],
        )
        with print_and_erase(PREPARING):
            self.analyzer = Analyzer.get(language)
        self.lemma_content_ids = defaultdict(set)
        self.form_content_ids = defaultdict(set)

    def update(self, exam: TextExam | DialogueExam, sentence: str) -> None:
        """Update the statistics with the given exam and sentence."""
        for word in self.analyzer.analyze(sentence):
            for stats_type, content_ids in [
                (StatsType.LEMMAS, self.lemma_content_ids[word.lemma]),
                (StatsType.FORMS, self.form_content_ids[get_normalized_form(word)]),
            ]:
                if exam.content_id not in content_ids:
                    content_ids.add(exam.content_id)
                    for period in get_relevant_periods(self.stats, exam):
                        update_stats(period, Pos.from_upos(word.upos), stats_type, len(content_ids))


def compute_stats(
        language: Language,
        paired: list[Language] | None,
        modality: Modality,
        checkpoints: list[datetime],
) -> PeriodStats:
    """Compute grammar statistics for the given language(s) and modality."""
    handler = StatsHandler(language, checkpoints)
    with single_session() as session:
        for exam in progress(
            merge(
                *[(item for item in session.query(model).filter(
                    get_filter(language, paired, modality, model),
                ).order_by(model.started_at))
                  for model in [TextExam, DialogueExam]],
                key=lambda item: item.started_at,
            ),
            total=sum(
                session.query(model).filter(get_filter(language, paired, modality, model)).count()
                for model in [TextExam, DialogueExam]
            ),
            modality=modality,
        ):
            for result, sentence in zip(
                    exam.results,
                    get_relevant_sentences(exam, language, modality),
            ):
                if result.score == MAX_SCORE:
                    handler.update(exam, sentence)
    return handler.stats


def format_increment(increment: int) -> str:
    return f'+{increment}' if increment > 0 else '0'


def build_stats_section(
        stats: PeriodStats,
        pos: Pos | None,
        stats_type: StatsType,
        threshold_headers: list[str],
        prepend_divider: bool = False,
) -> list[dict[str, str | Text]]:
    """Build the statistics section for the given part of speech and type."""
    pos_title = pos.to_table_title() if pos else TOTAL
    return [*([{
        POS: pos_title,
        TYPE: STATS_TYPE_DIVIDER,
        **{header: STATS_TYPE_DIVIDER for header in threshold_headers},
    }] if prepend_divider else []), {
        POS: pos_title,
        TYPE: stats_type.to_table_title(),
        **{header: str(get_pos_stats(stats.all_time, pos, stats_type, threshold))
           for header, threshold in zip(threshold_headers, REPETITION_THRESHOLDS)},
    }, *[{
        POS: pos_title,
        TYPE: colorize(since(checkpoint.start_time)),
        **{header: colorize(format_increment(get_pos_stats(checkpoint.stats, pos, stats_type, threshold)))
           for header, threshold in zip(threshold_headers, REPETITION_THRESHOLDS)},
    } for checkpoint in stats.checkpoints]]


def print_stats(stats: PeriodStats, modality: Modality, add_dividers: bool) -> None:
    """Print the provided statistics."""
    threshold_headers = [
        OCCURRENCES.format(occurrences=threshold) if threshold != 1 else ONCE
        for threshold in REPETITION_THRESHOLDS
    ]
    print_table(build_table(
        title=TITLE.format(modality=modality.to_table_title()),
        headers=[
            POS,
            TYPE,
            *threshold_headers,
        ],
        justify={
            POS: 'left',
            TYPE: 'left',
            **{header: 'right' for header in threshold_headers},
        },
        widths={
            header: max(len(line) for header in threshold_headers for line in header.splitlines())
            for header in threshold_headers
        },
        data=[row
              for pos in list(Pos) + [None]
              for stats_type in StatsType
              for row in build_stats_section(
                stats,
                pos,
                cast(StatsType, stats_type),
                threshold_headers,
                prepend_divider=add_dividers and stats_type != next(iter(StatsType)),
              )],
        group_by=[POS],
    ))


def display_stats(
        language: Language,
        paired: list[Language] | None,
        modality: Modality,
        checkpoints: list[datetime],
) -> None:
    """Display grammar statistics for the given language(s) and modality."""
    stats = compute_stats(language, paired, modality, checkpoints)
    print_stats(stats, modality, add_dividers=len(checkpoints) > 0)
